import torch
from torch.autograd     import grad
from shapes_toolbox.fidelities import kernel_distance, sinkhorn_distance, wasserstein_distance
from hausdorff    import hausdorff
from pykeops.torch         import Kernel

import numpy as np
from matplotlib import pyplot as plt
from scipy import misc
from time  import time

use_cuda = torch.cuda.is_available()
tensor   = torch.cuda.FloatTensor if use_cuda else torch.FloatTensor 

plt.ion()
plt.show()

s2v = lambda x : tensor([x])


from scipy.ndimage.filters import gaussian_filter

def LoadImage(fname) :
    img = misc.imread(fname, flatten = True) # Grayscale
    img = (img[::-1, :])  / 255.
    img = np.swapaxes(img, 0,1 )
    return tensor( 1 - img )

def extract_point_cloud(I, affine) :
    """Bitmap to point cloud."""

    # Threshold, to extract the relevant indices ---------------------------------------
    ind = (I > .001).nonzero()

    # Extract the weights --------------------------------------------------------------
    D = len(I.shape)
    if   D == 2 : mu_i = I[ind[:,0], ind[:,1]]
    elif D == 3 : mu_i = I[ind[:,0], ind[:,1], ind[:,2]]
    else : raise NotImplementedError()

    mu_i = mu_i * affine[0,0] * affine[1,1] 

    # Don't forget the changes of coordinates! -----------------------------------------
    M   = affine[:D,:D] ; off = affine[:D,D]
    x_i = ind.float() @ M.t() + off

    return ind, mu_i, x_i

def plot_point_cloud( mu, x, color, scale = 5.) :
    fig = plt.gcf()
    ax  = fig.gca()
    mu = mu.view(-1,1).detach().cpu().numpy()
    x  =  x.view(-1,2).detach().cpu().numpy()

    for (mu_i,x_i) in zip(mu, x) :
        circle_i = plt.Circle(x_i, scale*mu_i, color=color)
        ax.add_artist(circle_i)

# Load the pngs
dataset = "mini"
datasets = {
    "bars"     : ("data/hor_bars_1.png", "data/hor_bars_2.png"),
    "worms_c"  : ("data/worm_c.png",     "data/worm_d.png"),
    "mini"     : ("data/mini_a.png",     "data/mini_b.png"),
}
if dataset == "mini" :
    sample = 1
    smooth = False
    scale_grad = 0.01
else :
    sample = 4
    smooth = False
    scale_grad = 0.02

source = LoadImage(datasets[dataset][0])
target = LoadImage(datasets[dataset][1])

# The images are rescaled to fit into the unit square ==========================================
scale_i = source.shape[0]
affine = tensor( [[ 1, 0, 0 ],
                  [ 0, 1, 0 ] ]) / scale_i

params = {
    "formula"     : "kernel",
    "id"          : Kernel( "-distance(x,y)" ),
    "gamma"       : s2v( 1. ),
}

plt.figure(figsize=(10,10))

def test(name, params, bitmap=True, normalize=True, gradient=True, scale=3.) :

    params["kernel_heatmap_range"] = (0,1,100)

    ind_A, alpha_i, x_i = extract_point_cloud(source, affine)
    ind_B, beta_j,  y_j = extract_point_cloud(target, affine)

    if normalize :
        alpha_i = alpha_i / alpha_i.sum()
        beta_j  = beta_j  / beta_j.sum()

    x_i.requires_grad = True

    # Compute the distance between the *measures* A and B ------------------------------
    print("{:,}-by-{:,} KP: ".format(len(x_i), len(y_j)), end='')

    routines = { "hausdorff" :    hausdorff ,
                 "sinkhorn"     : sinkhorn_distance,
                 "wasserstein"  : wasserstein_distance,
                 "kernel" :       kernel_distance    }
    routine = routines[ params.get("formula", "hausdorff") ]

    cost, heatmaps = routine( params, (alpha_i,x_i), (beta_j,y_j), info = False )
    
    grad_x = grad( cost, [x_i] )[0] # gradient wrt the voxels' positions

    # Point cloud to bitmap (grad_x) ---------------------------------------------------
    tensor   = torch.cuda.FloatTensor if source.is_cuda else torch.FloatTensor 
    # Using torch.zero(...).dtype(cuda.FloatTensor) would be inefficient...
    # Let's directly make a "malloc", before zero-ing in place
    grad_src = tensor( *(tuple(source.shape) + (2,))  )
    grad_src.zero_()

    grad_src[ind_A[:,0],ind_A[:,1], :] = grad_x[:,:] 

    # Display ==================================================================================
    grad_src = - grad_src # N.B.: We want to visualize a descent direction, not the opposite!
    plt.clf()

    # Source + Target :
    if bitmap :
        source_plot = .3*source.cpu().numpy()
        target_plot = .3*target.cpu().numpy()
        img_plot    = np.dstack( ( np.ones(source_plot.shape) - target_plot, 
                                    np.ones(source_plot.shape) - source_plot - target_plot, 
                                    np.ones(source_plot.shape) - source_plot ) )
        
        plt.imshow( np.swapaxes(img_plot,0,1), origin="lower", extent=(0,1,0,1))
    else :
        plot_point_cloud(alpha_i, x_i+.5/scale_i, (.8,0.,0.), scale)
        plot_point_cloud(beta_j,  y_j+.5/scale_i, (0.,0.,.8), scale)


    if gradient :
        # Subsample the gradient field :
        grad_plot = grad_src.cpu().numpy()
        grad_plot = grad_plot[::sample, ::sample, :]
        if smooth :
            grad_plot = gaussian_filter(grad_plot, [1,1,0], mode='nearest')

        X,Y   = np.meshgrid( np.linspace(0, 1, grad_plot.shape[0]+1)[:-1] + .5/(sample*grad_plot.shape[0]), 
                            np.linspace(0, 1, grad_plot.shape[1]+1)[:-1] + .5/(sample*grad_plot.shape[1]) )
        U, V  = grad_plot[:,:,0], grad_plot[:,:,1]
        norms = np.sqrt( (grad_plot[:,:,1]**2 + grad_plot[:,:,0]**2) )
        mask  = (norms>0)
        scale = np.mean( norms[mask] )
        if smooth : mask  = (source_plot[::sample, ::sample]>0)

        Y, X, U, V = Y[mask], X[mask], U[mask], V[mask]
        plt.quiver( Y, X, U, V, 
                    scale = scale_grad*scale, scale_units="dots", color="#5CBF3A", zorder=3, width=0.01)

        # "Distance" fields :
        if heatmaps is not None :
            heatmaps.plot(plt.gca(), {})


    plt.axis([0,1,0,1])
    plt.axes().set_aspect('equal')
    # Save result in the "output/" folder :
    plt.savefig("output/bitmap/" + dataset + "_" + name +".png") 
    plt.pause(.01)


test("bitmap", params, bitmap=True,  normalize=False, gradient=False)
test("bitmap", params, bitmap=True,  normalize=False, gradient=False)
test("dots",   params, bitmap=False, normalize=False, gradient=False)
test("dots_n", params, bitmap=False, normalize=True,  gradient=False, scale=.3)
test("grad_n", params, bitmap=False, normalize=True,  gradient=True,  scale=.3)

print("Done.")
plt.show(block=True)
